@@ -1,7 +1,7 @@
 package Win32::FileOp;
 
 use vars qw($VERSION);
-$Win32::FileOp::VERSION = '0.10.4';
+$Win32::FileOp::VERSION = '0.11.3';
 
 use Win32::API;
 use File::Find;
@@ -62,8 +62,9 @@ my @CSIDL_flags = qw(
       OpenDialog SaveAsDialog BrowseForFolder
       recycle
       DesktopHandle GetDesktopHandle WindowHandle GetWindowHandle
-      Compress Uncompress UnCompress Compressed SetCompression GetCompression CompressDir UncompressDir UnCompressDir
+      Compress Uncompress UnCompress Compressed SetCompression GetCompression CompressedSize CompressDir UncompressDir UnCompressDir
       Map Unmap Disconnect Mapped
+      Subst Unsubst Substed SubstDev
  ),
  @FOF_flags,
  @OFN_flags,
@@ -72,7 +73,7 @@ my @CSIDL_flags = qw(
 );
 #     FOF_CONFIRMMOUSE FOF_WANTMAPPINGHANDLE
 
-@Win32::FileOp::EXPORT_OK = qw(FindInPATH FindInPath Relative2Absolute RelativeToAbsolute Map Unmap Disconnect Mapped);
+*Win32::FileOp::EXPORT_OK = \@Win32::FileOp::EXPORT;
 
 %Win32::FileOp::EXPORT_TAGS = (
     INI => [qw( ReadINISectionKeys ReadINISections WriteToINI WriteToWININI ReadINI ReadWININI DeleteFromINI DeleteFromWININI )],
@@ -90,8 +91,9 @@ my @CSIDL_flags = qw(
     _BASIC => [@FOF_flags],
     RECENT => [qw(AddToRecentDocs EmptyRecentDocs)],
     DIRECTORY => [qw(UpdateDir FillInDir)],
-    COMPRESS => [qw(Compress Uncompress UnCompress Compressed SetCompression GetCompression CompressDir UncompressDir UnCompressDir)],
+    COMPRESS => [qw(Compress Uncompress UnCompress Compressed SetCompression GetCompression CompressedSize CompressDir UncompressDir UnCompressDir)],
     MAP => [qw(Map Unmap Disconnect Mapped)],
+    SUBST => [qw(Subst Unsubst Substed SubstDev)]
 );
 
 
@@ -234,6 +236,12 @@ sub CSIDL_PRINTHOOD                 () { 0x001b }
 #sub TRUNCATE_EXISTING   () { 5 }
 #=cut
 
+sub DDD_RAW_TARGET_PATH         () { 0x00000001 }
+sub DDD_REMOVE_DEFINITION       () { 0x00000002 }
+sub DDD_EXACT_MATCH_ON_REMOVE   () { 0x00000004 }
+sub DDD_NO_BROADCAST_SYSTEM     () { 0x00000008 }
+
+
 #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
 tie $Win32::FileOp::fileop, 'Lazy', sub {
@@ -377,6 +385,12 @@ tie $Win32::FileOp::GetFileVersionInfo, 'Lazy', sub {
    die "new Win32::API::GetFileVersionInfo: $!\n"
 }, &LAZY_READONLY;
 
+tie $Win32::FileOp::GetCompressedFileSize, 'Lazy', sub {
+   new Win32::API("kernel32", "GetCompressedFileSize", ['P','P'], 'L')
+   or
+   die "new Win32::API::GetCompressedFileSize: $!\n"
+}, &LAZY_READONLY;
+
 tie $Win32::FileOp::VerQueryValue, 'Lazy', sub {
    new Win32::API( "version", "VerQueryValue", ['P', 'P', 'P', 'P'], 'N')
    or
@@ -407,6 +421,20 @@ tie $Win32::FileOp::GetLogicalDrives, 'Lazy', sub {
   die "new Win32::API::GetLogicalDrives: $!\n"
 }, &LAZY_READONLY;
 
+
+tie $Win32::FileOp::QueryDosDevice, 'Lazy', sub {
+  new Win32::API("kernel32.dll", "QueryDosDevice", ['P','P','L'], 'L')
+  or
+  die "new Win32::API::QueryDosDevice: $!\n"
+}, &LAZY_READONLY;
+
+tie $Win32::FileOp::DefineDosDevice, 'Lazy', sub {
+  new Win32::API("kernel32.dll", "DefineDosDevice", ['L','P','P'],'I')
+  or
+  die "new Win32::API::DefineDosDevice: $!\n"
+}, &LAZY_READONLY;
+
+
 #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
 sub Recycle {
@@ -482,7 +510,7 @@ sub _DeleteConfirmEach {
 
     my $res = 0;
     my ($filename,$opstruct);
-    while ($filename = shift) {
+    while (defined($filename = shift)) {
 
         Relative2Absolute $filename;
         $filename .= "\0\0";        # double term the filename
@@ -549,7 +577,7 @@ sub _MoveOrCopyEx {
 
     # fix to full paths
 
-        if (ref $from eq 'ARRAY') {
+        if (UNIVERSAL::isa($from, "ARRAY")) {
 
             my @files = map {
                 my $s = $_;
@@ -575,7 +603,7 @@ sub _MoveOrCopyEx {
         $from .= "\0\0";        # double term the filename
 
         my $options = $options;
-        if (ref $to eq 'ARRAY') {
+        if (UNIVERSAL::isa($to, "ARRAY")) {
             my $strto='';
             foreach (@$to) {
                 $strto .= RelativeToAbsolute($_) . "\0";
@@ -638,7 +666,7 @@ sub CopyFile {
     undef $Win32::FileOp::Error;
     my ($from,$to);
 
-    while ($from = shift and $to = shift) {
+    while (defined($from = shift) and defined($to = shift)) {
 #        Relative2Absolute($to,$from);
         $to .= "\0";
         $from .= "\0";
@@ -651,7 +679,7 @@ sub DeleteAtReboot {
     undef $Win32::FileOp::Error;
     if (Win32::IsWinNT)  {
         my $file;
-        while ($file = shift) {
+        while (defined($file = shift)) {
             Relative2Absolute($file);
             $Win32::FileOp::movefileexDel->Call($file, 0, MOVEFILE_DELAY_UNTIL_REBOOT);
         }
@@ -673,7 +701,7 @@ sub MoveFileEx {
     my $options = pop;
 
     my ($from,$to);
-    while ($from = shift and $to = shift) {
+    while (defined($from = shift) and defined($to = shift)) {
         Relative2Absolute($to,$from);
         $to .= "\0";
         $from .= "\0";
@@ -749,7 +777,7 @@ sub AddToRecentDocs {
 
  my $file;
  my $res=0;
- while ($file = shift) {
+ while (defined($file = shift)) {
   next unless -e $file;
   Relative2Absolute($file);
   $file .= "\0";
@@ -772,7 +800,7 @@ sub WriteToINI {
     my $INI = shift;$INI .= "\0";
     my $section = shift;$section .= "\0";
     my ($name,$value);
-    while ($name = shift and $value = shift) {
+    while (defined($name = shift) and defined($value = shift)) {
         $name .= "\0";$value .= "\0";
         $Win32::FileOp::writeINI->Call($section,$name,$value,$INI)
         or return undef;
@@ -784,7 +812,7 @@ sub WriteToWININI {
     undef $Win32::FileOp::Error;
     my $section = shift;$section .= "\0";
     my ($name,$value);
-    while ($name = shift and $value = shift) {
+    while (defined($name = shift) and defined($value = shift)) {
         $name .= "\0";$value .= "\0";
         $Win32::FileOp::writeWININI->Call($section,$name,$value)
         or return undef;
@@ -797,7 +825,7 @@ sub DeleteFromINI {
     my $INI = shift;$INI .= "\0";
     my $section = shift;$section .= "\0";
     my $name;
-    while ($name = shift) {
+    while (defined($name = shift)) {
         $name .= "\0";
         $Win32::FileOp::deleteINI->Call($section,$name,0,$INI)
         or return undef;
@@ -809,7 +837,7 @@ sub DeleteFromWININI {
     undef $Win32::FileOp::Error;
     my $section = shift;$section .= "\0";
     my $name;
-    while ($name = shift) {
+    while (defined($name = shift)) {
         $name .= "\0";
         $Win32::FileOp::deleteWININI->Call($section,$name,0)
         or return undef;
@@ -839,7 +867,7 @@ sub ReadINISectionKeys {
 	my @values;
 
 	@values = split(/\0/,_ReadINI($section,$name,$default,$INI));
-	@{$_[0]} = @values if (ref $_[0] eq 'ARRAY');
+	@{$_[0]} = @values if (UNIVERSAL::isa($_[0], "ARRAY"));
 	return wantarray() ? @values : (@values ? \@values : undef);
 }
 # END MTY Hack
@@ -853,7 +881,7 @@ sub ReadINISections {
 	my @values;
 
 	@values = split(/\0/,_ReadINI($section,$name,$default,$INI));
-	@{$_[0]} = @values if (ref $_[0] eq 'ARRAY');
+	@{$_[0]} = @values if (UNIVERSAL::isa($_[0], "ARRAY"));
 	return wantarray() ? @values : (@values ? \@values : undef);
 }
 
@@ -898,7 +926,7 @@ sub OpenOrSaveDialog {
     undef $Win32::FileOp::Error;
     my $fun = shift;
     my $params;
-    if ((ref $_[0]) =~ /HASH/) {
+    if (UNIVERSAL::isa($_[0], "HASH")) {
         $params = $_[0];
         $params->{filename} = $_[1] if defined $_[1];
     } else {
@@ -910,15 +938,18 @@ sub OpenOrSaveDialog {
             $params = {@_};
         }
     }
+    foreach (grep {s/^-//} keys %$params) {$params->{$_} = $params->{"-$_"};delete $params->{"-$_"}};
 
     $params->{handle} = 'self' unless exists $params->{handle};
+    $params->{options} = 0 unless exists $params->{options};
+    
 
     my $lpstrFilter = '';
-    if ((ref $params->{filters}) =~ /HASH/) {
+    if (UNIVERSAL::isa($params->{filters}, "HASH")) {
         foreach (keys %{$params->{filters}}) {
             $lpstrFilter .= $_ . "\0" . $params->{filters}->{$_} . "\0";
         }
-    } elsif ((ref $params->{filters}) =~ /ARRAY/) {
+    } elsif (UNIVERSAL::isa($params->{filters}, "ARRAY")) {
         my ($title,$filter,$i);
         $i=0;$lpstrFilter='';
         while ($title = ${$params->{filters}}[$i++] and $filter = ${$params->{filters}}[$i++]) {
@@ -932,8 +963,7 @@ sub OpenOrSaveDialog {
         $lpstrFilter = "\0\0";
     }
 
-my $W = $^W;
-$^W=0;
+local $^W = 0;
 
     my $nFilterIndex = $params->{defaultfilter};
     $nFilterIndex = 1 unless $nFilterIndex>0; # to be sure it's a reasonable number
@@ -948,14 +978,14 @@ $^W=0;
     my $lpstrInitialDir = $params->{dir} . "\0";
     my $lpstrTitle  = $params->{title} . "\0";
     my $Flags = $params->{options};
+    my $nFileExtension = "\0\0";
     my $lpstrDefExt = $params->{extension}."\0";
     my $lpTemplateName = "\0";
     my $Handle = $params->{handle};
     if ($Handle =~ /^self$/i) {$Handle = GetWindowHandle()};
-    
-$^W=$W;
 
-    my $struct = pack "LLLpLLLpLpLppLIIpLLp",
+#    my $struct = pack "LLLpLLLpLpLppLIIpLLp",
+    my $struct = pack "LLLpLLLpLpLppLIppLLp",
      (
       76,                        #'lStructSize'       #  DWORD
       $Handle,                   #'hwndOwner'         #  HWND
@@ -976,7 +1006,8 @@ $^W=$W;
       $lpstrTitle,               #'lpstrTitle'        #  LPCTSTR
       $Flags,                    #'Flags'             #  DWORD
       0,                         #'nFileOffset'       #  WORD
-      0,                         #'nFileExtension'    #  WORD
+#      0,                         #'nFileExtension'    #  WORD
+      $nFileExtension,           #'nFileExtension'    #  WORD
       $lpstrDefExt,              #'lpstrDefExt'       #  LPCTSTR
       0,                         #'lCustData'         #  DWORD
       0,                         #'lpfnHook'          #  LPOFNHOOKPROC
@@ -1025,7 +1056,7 @@ sub BrowseForFolder {
       =
       (GetWindowHandle(), "\0"x260, " "x260,
        shift() || "\0", shift() || 0, (shift() || 0) | 0x0000,
-       "\0", 0, 0, " "x260);
+       0, 0, 0, "\0"x260);
 
    $nFolder = CSIDL_DRIVES() unless defined $nFolder;
 
@@ -1034,7 +1065,7 @@ sub BrowseForFolder {
    $pidlRoot =~ s/\0.*$//s;
    $pidlRoot = hex unpack 'H*',(join'', reverse split//, $pidlRoot);
 
-   my $browseinfo = pack 'LLppIaLI',
+   my $browseinfo = pack 'LLppILLI',
       ($hwndOwner, $pidlRoot, $pszDisplayName, $lpszTitle,
        $ulFlags, $lpfn, $lParam, $iImage);
 
@@ -1087,7 +1118,7 @@ sub GetWindowHandle {
     if (! $Win32::FileOp::WindowHandle) {
         my $GetConsoleTitle = new Win32::API("kernel32", "GetConsoleTitle", ['P','N'],'N');
         my $SetConsoleTitle = new Win32::API("kernel32", "SetConsoleTitle", ['P'],'N');
-        my $SleepEx = new Win32::API("kernel32", "Sleep", ['N','I'],'V');
+        my $SleepEx = new Win32::API("kernel32", "SleepEx", ['N','I'],'V');
         my $FindWindow = new Win32::API("user32", "FindWindow", ['P','P'],'N');
 
         my $oldtitle = " " x 1024;
@@ -1113,7 +1144,7 @@ sub SetCompression {
         $flag = 1;
     }
     $_[0] = $_ unless @_;
-    while ($file = shift) {
+    while (defined($file = shift)) {
 
 #print "\t$file\n";
 
@@ -1176,6 +1207,12 @@ sub Uncompress {SetCompression(@_,0)}
 *UnCompress = \&Uncompress;
 sub Compressed {&GetCompression}
 
+sub CompressedSize {
+ my $file = $_[0];
+ my $hsize = "\0" x 4;
+ my $lsize = $Win32::FileOp::GetCompressedFileSize->Call( $file, $hsize);
+ return $lsize + 0x10000*unpack('L',$hsize);
+}
 
 sub UncompressDir {
     undef $Win32::FileOp::Error;
@@ -1220,7 +1257,7 @@ sub Map {
  $share .= "\0";
  my $opt = shift || {};
  croak 'Ussage: Win32::FileOp::Map([$drive,]$share[,\%options])',"\n"
-  unless (ref $opt eq 'HASH');
+  unless (UNIVERSAL::isa($opt, "HASH"));
  my $username = 0;
  if (defined $opt->{user}) {
   $username = $opt->{user}."\0";
@@ -1255,7 +1292,7 @@ sub Disconnect {
  my $disk = shift() . "\0";$disk =~ s/^(\w)\0$/$1:\0/;
  my $opt = shift() || {};
  croak 'Ussage: Win32::FileOp::Map([$drive,]$share[,\%options])',"\n"
-  unless (ref $opt eq 'HASH');
+  unless (UNIVERSAL::isa($opt, "HASH"));
  my $options = $opt->{persistent} ? 1 : 0;
  my $force   = $opt->{force} ? 1 : 0;
 
@@ -1266,7 +1303,7 @@ sub Disconnect {
 
 sub Unmap {
     undef $Win32::FileOp::Error;
-    if (ref $_[1] eq 'HASH') {
+    if (UNIVERSAL::isa($_[1], "HASH")) {
         $_[1]->{persistent} = 1 unless exists $_[1]->{persistent};
     } else {
         $_[1] = {persistent => 1}
@@ -1311,6 +1348,63 @@ sub _MappedAll {
     return %hash;
 }
 
+#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+
+sub Subst {
+ my $drive = shift;
+ return unless $drive =~ s/^(\w):?$/$1:\0/;
+ my $path = shift();
+ return unless -e $path;
+ $path.="\0";
+ $Win32::FileOp::DefineDosDevice->Call(0,$drive,$path);
+}
+
+sub SubstDev {
+ my $drive = shift;
+ return unless $drive =~ s/^(\w):?$/$1:\0/;
+ my $path = shift();
+# return unless -e $path;
+ $path = "\\Device\\$path" unless $path =~ /\\Device\\/i;
+ $path.="\0";
+ $Win32::FileOp::DefineDosDevice->Call(&DDD_RAW_TARGET_PATH,$drive,$path);
+}
+
+sub Unsubst {
+ my $drive = shift;
+ return unless $drive =~ s/^(\w):?$/$1:\0/;
+ $Win32::FileOp::DefineDosDevice->Call(&DDD_REMOVE_DEFINITION,$drive,0);
+}
+
+sub Substed {
+ my $drive = shift;
+ if (defined $drive) {
+  return unless $drive =~ s/^(\w):?$/$1:\0/;
+  my $path = "\0" x 1024;
+  my $device;
+  $Win32::FileOp::QueryDosDevice->Call($drive,$path,1024)
+   or return;
+
+  $path =~ s/\0.*$//;
+  
+  $path =~ s/^\\\?\?\\UNC/\\/ and $device = 'UNC'
+  or
+  $path =~ s/\\Device\\(.*?)\\\w:/\\/ and $device = $1
+  or
+  $path =~ s/\\Device\\(.*)$// and $device = $1;
+  
+  return wantarray ? ($path,$device) : $path;
+ } else {
+  my ($drive,$path,%data);
+  foreach $drive (('A'..'Z')) {
+    $drive.=':';
+    $path = Substed($drive);
+    $data{$drive} = $path if defined $path;
+  }
+  return wantarray() ? %data : \%data;
+ }
+}
+
+
 #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
 package Win32::FileOp::Error;
@@ -1340,16 +1434,16 @@ __END__
 
 =head1 NAME
 
- Win32::FileOp - 0.10.3
+ Win32::FileOp - 0.11.3
 
 =head1 DESCRIPTION
 
 Module for file operations with fancy dialog boxes, for moving files
 to recycle bin, reading and updating INI files and file operations in general.
 
-Unless mentioned otherwise all functions work both under WinNT and Win95.
+Unless mentioned otherwise all functions work both under WinNT and Win9x.
 
-Version 0.10.3
+Version 0.11.3
 
 =head2 Functions
 
@@ -1380,6 +1474,11 @@ C<OpenDialog> C<SaveAsDialog> C<BrowseForFolder>
 
 C<Map> C<Unmap> C<Disconnect> C<Mapped>
 
+C<Subst> C<Unsubst> <Substed>
+
+To get the error message from most of these functions, you should not use $!, but
+$^E or Win32::FormatMessage(Win32::GetLastError())!
+
 =over 2
 
 =item GetDesktopHandle
@@ -1811,10 +1910,6 @@ It also sets two global variables :
                    ordering!
   dir => the initial directory for the dialog, default is the current directory
   filename => the default filename to be showed in the dialog
-  extension => the default extension, this function appends this extension to
-               the filename if the user fails to type an extension. This
-               string can be any length, but only the first three characters
-               are appended. The string should not contain a period (.).
   handle => the handle to the window which will own this dialog
             Default is the console of the perl script.
             If you do not want to tie the dialog to any window use
@@ -1835,6 +1930,8 @@ REM: Based on Win32 API function GetOpenFileName().
 
 Creates the Save As dialog box, parameters are the same as for OpenDialog.
 
+REM: Based on Win32 API function GetSaveFileName().
+
 =item BrowseForFolder
 
  BrowseForFolder [$title [, $rootFolder [, $options]]]
@@ -1845,7 +1942,7 @@ The $rootFolder may be one of the C<CSIDL_>... constants.
 For $options you should use the C<BIF_>... constants. Description
 of the constants is bellow.
 
-REM: Based on Win32 API function GetSaveFileName().
+REM: Based on Win32 API function SHBrowseForFolder().
 
 =item Map
 
@@ -1923,6 +2020,78 @@ Win32::FormatMessage(Win32::GetLastError())!
 
 REM: Based on Win32 API function WNetGetConnection().
 
+
+=item Subst
+
+ Subst Z => 'c:\temp';
+ Subst 'Z:' => '\\\\servername\share\subdir';
+
+This function substitutes a drive letter for a directory, both local and UNC.
+
+Be very carefull with this, cause it'll allow you to change the
+substitution even for C:. ! Which will most likely be lethal !
+
+Works only on WinNT.
+
+REM: Based on DefineDosDevice()
+
+ =item SubstDev
+
+ SubstDev F => 'Floppy0';
+ SubstDev G => 'Harddisk0\Partition1';
+
+Allows you to make a substitution to devices. For example if you want to
+make an alias for A: ...
+
+To get the device mapped to a drive letter use Substed() in list context.
+
+Works only on WinNT.
+
+REM: Based on DefineDosDevice()
+
+=item Unsubst
+
+ Unsubst 'X';
+
+Deletes the substitution for a drive letter. Again, be very carefull with this!
+
+Works only on WinNT.
+
+REM: Based on DefineDosDevice()
+
+=item Substed
+
+ %drives = Substed;
+ $substitution = Substed $drive;
+ ($substitution, $device) = Substed $drive;
+
+This function retrieves the name of the resource(s) associated with a drive letter(s).
+
+If used with a parameter :
+
+In scalar context you get the substitution. If the drive is the root of
+a local device you'll get an empty string, if it's not mapped to
+anything you'll get undef.
+
+In list context you'll get both the substitution and the device/type of device :
+
+ Substed 'A:' => ('','Floppy0')
+ Substed 'B:' => undef
+ Substed 'C:' => ('','Harddisk0\Partition1')
+ Substed 'H:' => ('\\\\servername\homes\username','UNC')
+  # set by subst H: \\servername\homes\username
+ Substed 'S:' => ('\\\\servername\servis','LanmanRedirector')
+  # set by net use S: \\servername\servis
+ Substed 'X:' => ()
+  # not mapped to anything
+
+If used without a parameter gives you a hash of drives and their
+corresponding sunstitutions.
+
+Works only on WinNT.
+
+REM: Based on Win32 API function QueryDosDevice().
+
 =back
 
 =head2 Options
@@ -2244,13 +2413,17 @@ The available groups are :
  BASIC = Move..., Copy..., Recycle... and Delete... functions plus constants
  _BASIC = FOF_... constants only
  HANDLES = DesktopHandle GetDesktopHandle WindowHandle GetWindowHandle
- INI = WriteToINI WriteToWININI ReadINI ReadWININI ReadINISectionKeys DeleteFromINI DeleteFromWININI
- DIALOGS = OpenDialog SaveAsDialog plus OFN_... constants
- _DIALOGS = only OFN_... constants
+ INI = WriteToINI WriteToWININI ReadINI ReadWININI ReadINISectionKeys
+       DeleteFromINI DeleteFromWININI
+ DIALOGS = OpenDialog, SaveAsDialog and BrowseForFolder plus OFN_...,
+           BIF_... and CSIDL_... constants
+ _DIALOGS = only OFN_..., BIF_... and CSIDL_... constants
  RECENT = AddToRecentDocs, EmptyRecentDocs
  DIRECTORY = UpdateDir, FillInDir
  COMPRESS => Compress Uncompress Compressed SetCompression GetCompression
+             CompressedSize CompressDir UncompressDir
  MAP => Map Unmap Disconnect Mapped
+ SUBST => Subst Unsubst Substed SubstDev
 
 Examples:
 
